import json
import logging
import shutil
import sys
from glob import glob
from os import listdir, path, remove, rmdir
from pathlib import Path

import pytest

import mimeo.__main__ as MimeoCLI


@pytest.fixture(autouse=True)
def minimum_config():
    return {
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True
                    }
                }
            }
        ]
    }


@pytest.fixture(autouse=True)
def default_config():
    return {
        "output_format": "xml",
        "indent": 4,
        "xml_declaration": True,
        "output_details": {
            "direction": "file",
            "directory_path": "test_mimeo_cli-dir/output",
            "file_name": "output-file"
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True
                    }
                }
            }
        ]
    }


@pytest.fixture(autouse=True)
def setup_and_teardown(default_config, minimum_config):
    # Setup
    Path("test_mimeo_cli-dir").mkdir(parents=True, exist_ok=True)
    with open("test_mimeo_cli-dir/default-config.json", "w") as file:
        json.dump(default_config, file)
    with open("test_mimeo_cli-dir/minimum-config.json", "w") as file:
        json.dump(minimum_config, file)

    yield

    # Teardown
    shutil.rmtree("test_mimeo_cli-dir")
    for filename in glob("mimeo-output/customized-output-file*"):
        remove(filename)
    if path.exists("mimeo-output") and not listdir("mimeo-output"):
        rmdir("mimeo-output")


def test_basic_use():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_short_xml_declaration_false():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-x", "false"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_short_xml_declaration_true():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-x", "true"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_xml_declaration_false():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--xml-declaration", "false"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_xml_declaration_true():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--xml-declaration", "true"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_short_indent_non_zero():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-i", "2"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '  <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '  <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '  <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_short_indent_zero():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-i", "0"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == "<?xml version='1.0' encoding='utf-8'?>\n"
            assert file_content.readline() == '<SomeEntity>' \
                                              '<ChildNode1>1</ChildNode1>' \
                                              '<ChildNode2>value-2</ChildNode2>' \
                                              '<ChildNode3>true</ChildNode3>' \
                                              '</SomeEntity>'


def test_custom_long_indent_non_zero():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--indent", "2"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '  <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '  <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '  <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_indent_zero():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--indent", "0"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == "<?xml version='1.0' encoding='utf-8'?>\n"
            assert file_content.readline() == '<SomeEntity>' \
                                              '<ChildNode1>1</ChildNode1>' \
                                              '<ChildNode2>value-2</ChildNode2>' \
                                              '<ChildNode3>true</ChildNode3>' \
                                              '</SomeEntity>'


def test_custom_short_output_direction():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-o", "stdout"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert not path.exists("test_mimeo_cli-dir/output")


def test_custom_long_output_direction():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--output", "stdout"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert not path.exists("test_mimeo_cli-dir/output")


def test_custom_output_direction_does_not_throw_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "stdout"]

    try:
        MimeoCLI.main()
    except KeyError:
        assert False


def test_custom_short_output_directory_path():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-d", "test_mimeo_cli-dir/customized-output"]

    assert not path.exists("test_mimeo_cli-dir/output")
    assert not path.exists("test_mimeo_cli-dir/customized-output")

    MimeoCLI.main()

    assert not path.exists("test_mimeo_cli-dir/output")
    assert path.exists("test_mimeo_cli-dir/customized-output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/customized-output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_output_directory_path():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--directory", "test_mimeo_cli-dir/customized-output"]

    assert not path.exists("test_mimeo_cli-dir/output")
    assert not path.exists("test_mimeo_cli-dir/customized-output")

    MimeoCLI.main()

    assert not path.exists("test_mimeo_cli-dir/output")
    assert path.exists("test_mimeo_cli-dir/customized-output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/customized-output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_output_directory_path_does_not_throw_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-d", "test_mimeo_cli-dir/customized-output"]

    try:
        MimeoCLI.main()
    except KeyError:
        assert False


def test_custom_short_output_file_name():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-f", "customized-output-file"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/customized-output-file-{i}.xml"
        assert path.exists(file_path)
        assert not path.exists(f"test_mimeo_cli-dir/output/output-file-{i}.xml")

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_output_file_name():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--file", "customized-output-file"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/customized-output-file-{i}.xml"
        assert path.exists(file_path)
        assert not path.exists(f"test_mimeo_cli-dir/output/output-file-{i}.xml")

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_output_file_name_does_not_throw_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-f", "customized-output-file"]

    try:
        MimeoCLI.main()
    except KeyError:
        assert False


def test_logging_mode_default():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json"]
    logger = logging.getLogger("mimeo")

    MimeoCLI.main()

    assert logger.getEffectiveLevel() == logging.INFO


def test_logging_mode_silent():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--silent"]
    logger = logging.getLogger("mimeo")

    MimeoCLI.main()

    assert logger.getEffectiveLevel() == logging.WARNING


def test_logging_mode_debug():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--debug"]
    logger = logging.getLogger("mimeo")

    MimeoCLI.main()

    assert logger.getEffectiveLevel() == logging.DEBUG


def test_logging_mode_fine():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--fine"]
    logger = logging.getLogger("mimeo")

    MimeoCLI.main()

    assert logger.getEffectiveLevel() == logging.FINE



class NotAllowedInstantiation(Exception):
    pass


class InvalidMimeoUtil(Exception):
    pass


class NotASpecialField(Exception):
    pass


class InvalidSpecialFieldValue(Exception):
    pass


class CountryNotFound(Exception):
    pass


class OutOfStock(Exception):
    pass



from mimeo.config.mimeo_config import MimeoConfig, UnsupportedOutputFormat
from mimeo.generators import Generator, GeneratorUtils, XMLGenerator


class GeneratorFactory:

    XML = "xml"

    @staticmethod
    def get_generator(mimeo_config: MimeoConfig) -> Generator:
        output_format = mimeo_config.output_format
        if output_format == GeneratorFactory.XML:
            return XMLGenerator(mimeo_config)
        else:
            raise UnsupportedOutputFormat(f"Provided format [{output_format}] is not supported!")



from __future__ import annotations

import datetime
import random
import re
import string
import uuid
from datetime import date, datetime, timedelta

from mimeo.config import MimeoConfig
from mimeo.database import Country, MimeoDB
from mimeo.generators.exc import (CountryNotFound, InvalidMimeoUtil,
                                  InvalidSpecialFieldValue,
                                  NotAllowedInstantiation, NotASpecialField,
                                  OutOfStock)


class GeneratorUtils:

    __CREATE_KEY = object()
    __MIMEO_DB = MimeoDB()
    __INSTANCES = {}
    __VARS = {}

    @classmethod
    def setup(cls, mimeo_config: MimeoConfig) -> None:
        cls.__VARS = mimeo_config.vars

    @classmethod
    def get_for_context(cls, context: str) -> GeneratorUtils:
        if context not in GeneratorUtils.__INSTANCES:
            cls.__INSTANCES[context] = GeneratorUtils(cls.__CREATE_KEY)
        return cls.__INSTANCES[context]

    def __init__(self, create_key):
        GeneratorUtils.__validate_instantiation(create_key)
        self.__id = 0
        self.__curr_iter = 0
        self.__keys = []
        self.__special_fields = {}
        self.__cities_indexes = {}
        self.__countries_indexes = {}

    def reset(self) -> None:
        self.__id = 0

    def setup_iteration(self, curr_iter: int) -> None:
        self.__curr_iter = curr_iter
        self.__keys.append(str(uuid.uuid4()))
        self.__special_fields = {}

    def auto_increment(self, pattern="{:05d}") -> str:
        try:
            self.__id += 1
            return pattern.format(self.__id)
        except AttributeError as err:
            self.__id -= 1
            raise err

    def curr_iter(self, context: str = None) -> int:
        if context is not None:
            return GeneratorUtils.get_for_context(context).curr_iter()
        return self.__curr_iter

    def key(self) -> str:
        return self.__keys[-1]

    def city(self, allow_duplicates: bool = False) -> str:
        if allow_duplicates:
            return self.__MIMEO_DB.get_city_at(self.random_int(MimeoDB.NUM_OF_CITIES)).name_ascii
        else:
            self.__initialize_cities_indexes("_ALL_", MimeoDB.NUM_OF_CITIES)

            if len(self.__cities_indexes["_ALL_"]) == 0:
                raise OutOfStock(f"No more unique values, database contain only {MimeoDB.NUM_OF_CITIES} cities.")
            else:
                index = self.__cities_indexes["_ALL_"].pop()
                city = self.__MIMEO_DB.get_city_at(index)
                return city.name_ascii

    def city_of(self, country: str, allow_duplicates: bool = False) -> str:
        country_cities = self.__MIMEO_DB.get_cities_of(country)
        country_cities_count = len(country_cities)
        if country_cities_count == 0:
            raise CountryNotFound(f"Mimeo database does not contain any cities of provided country [{country}].")
        elif allow_duplicates:
            return country_cities[self.random_int(country_cities_count)].name_ascii
        else:
            self.__initialize_cities_indexes(country, country_cities_count)

            if len(self.__cities_indexes[country]) == 0:
                raise OutOfStock(f"No more unique values, database contain only {country_cities_count} cities of {country}.")
            else:
                index = self.__cities_indexes[country].pop()
                city = country_cities[index]
                return city.name_ascii

    def country(self, allow_duplicates: bool = False, country: str = None) -> str:
        return self.__get_country(allow_duplicates, country).name

    def country_iso3(self, allow_duplicates: bool = False, country: str = None) -> str:
        return self.__get_country(allow_duplicates, country).iso_3

    def country_iso2(self, allow_duplicates: bool = False, country: str = None) -> str:
        return self.__get_country(allow_duplicates, country).iso_2

    def provide(self, field_name: str, field_value) -> None:
        if not GeneratorUtils.is_special_field(field_name):
            raise NotASpecialField(f"Provided field [{field_name}] is not a special one (use {'{:NAME:}'})!")
        if isinstance(field_value, dict) or isinstance(field_value, list):
            raise InvalidSpecialFieldValue(f"Provided field value [{field_value}] is invalid (use any atomic value)!")

        self.__special_fields[field_name] = field_value

    def inject(self, field_name: str):
        if field_name not in self.__special_fields:
            raise NotASpecialField(f"There's no such a special field [{field_name[2:][:-2]}]!")
        return self.__special_fields.get(field_name)

    def __get_country(self, allow_duplicates: bool = False, country: str = None) -> Country:
        if country is not None:
            countries = self.__MIMEO_DB.get_countries()
            country_found = next(filter(lambda c: c.name == country or c.iso_3 == country or c.iso_2 == country, countries), None)
            if country_found is not None:
                return country_found
            else:
                raise CountryNotFound(f"Mimeo database does not contain such a country [{country}].")
        elif allow_duplicates:
            return self.__MIMEO_DB.get_country_at(self.random_int(MimeoDB.NUM_OF_COUNTRIES))
        else:
            self.__initialize_countries_indexes("_ALL_", MimeoDB.NUM_OF_COUNTRIES)

            if len(self.__countries_indexes["_ALL_"]) == 0:
                raise OutOfStock(f"No more unique values, database contain only {MimeoDB.NUM_OF_COUNTRIES} countries.")
            else:
                index = self.__countries_indexes["_ALL_"].pop()
                country = self.__MIMEO_DB.get_country_at(index)
                return country

    def __initialize_cities_indexes(self, key: str, number_of_entries: int):
        if key not in self.__cities_indexes:
            self.__cities_indexes[key] = random.sample(range(number_of_entries), number_of_entries)

    def __initialize_countries_indexes(self, key: str, number_of_entries: int):
        if key not in self.__countries_indexes:
            self.__countries_indexes[key] = random.sample(range(number_of_entries), number_of_entries)

    @staticmethod
    def get_key(context: str, iteration: int = 0) -> str:
        return GeneratorUtils.get_for_context(context).__keys[iteration - 1]

    @staticmethod
    def random_str(length=20) -> str:
        return "".join(random.choice(string.ascii_letters) for _ in range(length))

    @staticmethod
    def random_int(limit: int = 100) -> int:
        return random.randrange(limit)

    @staticmethod
    def random(items: list = None):
        items = items if items is not None else []
        length = len(items)
        return "" if length == 0 else items[random.randrange(0, len(items))]

    @staticmethod
    def date(days_delta=0) -> str:
        date_value = date.today() if days_delta == 0 else date.today() + timedelta(days=days_delta)
        return date_value.strftime("%Y-%m-%d")

    @staticmethod
    def date_time(days_delta=0, hours_delta=0, minutes_delta=0, seconds_delta=0) -> str:
        time_value = datetime.now() + timedelta(days=days_delta,
                                                hours=hours_delta,
                                                minutes=minutes_delta,
                                                seconds=seconds_delta)
        return time_value.strftime("%Y-%m-%dT%H:%M:%S")

    @staticmethod
    def get_special_field_name(field_name: str) -> str:
        if not GeneratorUtils.is_special_field(field_name):
            raise NotASpecialField(f"Provided field [{field_name}] is not a special one (use {'{:NAME:}'})!")

        return field_name[2:][:-2]

    @staticmethod
    def is_special_field(field_name: str) -> bool:
        return bool(re.match(r"^{:([a-zA-Z]+:)?[-_a-zA-Z0-9]+:}$", field_name))

    @staticmethod
    def render_value(context: str, value):
        value_str = str(value)
        if isinstance(value, bool):
            return value_str.lower()

        special_fields_pattern = re.compile(".*({:([a-zA-Z]+:)?[-_a-zA-Z0-9]+:})")
        vars_pattern = re.compile(".*({[A-Z_0-9]+})")
        funct_pattern = re.compile("^{(.+)}$")
        try:
            if special_fields_pattern.match(value_str):
                match = next(special_fields_pattern.finditer(value_str))
                mimeo_util = match.group(1)
                rendered_value = GeneratorUtils.get_for_context(context).inject(mimeo_util)
                final_rendered_value = value_str.replace(mimeo_util, str(rendered_value))
                return GeneratorUtils.render_value(context, final_rendered_value)
            elif vars_pattern.match(value_str):
                match = next(vars_pattern.finditer(value_str))
                mimeo_util = match.group(1)
                rendered_value = GeneratorUtils.__render_var(context, mimeo_util[1:][:-1])
                final_rendered_value = value_str.replace(mimeo_util, str(rendered_value))
                return GeneratorUtils.render_value(context, final_rendered_value)
            elif funct_pattern.match(value_str):
                match = next(funct_pattern.finditer(value_str))
                mimeo_util = match.group(1)
                rendered_value = GeneratorUtils.__eval_funct(context, mimeo_util)
                rendered_value_str = str(rendered_value)
                if isinstance(rendered_value, bool):
                    return rendered_value_str.lower()
                return rendered_value_str
        except InvalidMimeoUtil:
            pass
        return value_str

    @staticmethod
    def __render_var(context: str, mimeo_util: str):
        value = GeneratorUtils.__VARS.get(mimeo_util)
        if value is not None:
            return GeneratorUtils.render_value(context, value)
        else:
            raise InvalidMimeoUtil(f"Provided variable [{mimeo_util}] is not defined!")

    @staticmethod
    def __eval_funct(context: str, funct: str):
        utils = GeneratorUtils.get_for_context(context)
        prepared_funct = funct
        prepared_funct = re.sub(r"auto_increment\((.*)\)", r"utils.auto_increment(\1)", prepared_funct)
        prepared_funct = re.sub(r"curr_iter\((.*)\)", r"utils.curr_iter(\1)", prepared_funct)
        if "get_key" in prepared_funct:
            prepared_funct = re.sub(r"get_key\((.*)\)", r"utils.get_key(\1)", prepared_funct)
        elif "key" in prepared_funct:
            prepared_funct = re.sub(r"key\((.*)\)", r"utils.key(\1)", prepared_funct)
        prepared_funct = re.sub(r"random_str\((.*)\)", r"utils.random_str(\1)", prepared_funct)
        prepared_funct = re.sub(r"random_int\((.*)\)", r"utils.random_int(\1)", prepared_funct)
        prepared_funct = re.sub(r"random\((.*)\)", r"utils.random(\1)", prepared_funct)
        prepared_funct = re.sub(r"date\((.*)\)", r"utils.date(\1)", prepared_funct)
        prepared_funct = re.sub(r"date_time\((.*)\)", r"utils.date_time(\1)", prepared_funct)
        prepared_funct = re.sub(r"city\((.*)\)", r"utils.city(\1)", prepared_funct)
        prepared_funct = re.sub(r"city_of\((.*)\)", r"utils.city_of(\1)", prepared_funct)
        prepared_funct = re.sub(r"country\((.*)\)", r"utils.country(\1)", prepared_funct)
        prepared_funct = re.sub(r"country_iso3\((.*)\)", r"utils.country_iso3(\1)", prepared_funct)
        prepared_funct = re.sub(r"country_iso2\((.*)\)", r"utils.country_iso2(\1)", prepared_funct)
        if prepared_funct.startswith("utils"):
            try:
                return eval(prepared_funct)
            except (TypeError, AttributeError, SyntaxError) as e:
                pass
        raise InvalidMimeoUtil(f"Provided function [{funct}] is invalid!")

    @staticmethod
    def __validate_instantiation(create_key: str) -> None:
        try:
            assert (create_key == GeneratorUtils.__CREATE_KEY)
        except AssertionError:
            raise NotAllowedInstantiation("GeneratorUtils cannot be instantiated directly! "
                                          "Please use GeneratorUtils.get_for_context(context)")



import json
import logging
from argparse import ArgumentParser

from mimeo import Mimeograph
from mimeo.config import MimeoConfig
from mimeo.logging import setup_logging

setup_logging()
logger = logging.getLogger(__name__)


class MimeoArgumentParser(ArgumentParser):

    def __init__(self):
        super().__init__(
            prog="mimeo",
            description="Generate data based on a template")
        self.add_argument(
            "-v",
            "--version",
            action="version",
            version="%(prog)s v0.1.1")
        self.add_argument(
            "paths",
            nargs="+",
            type=str,
            help="take paths to Mimeo Configurations")

        mimeo_config_args = self.add_argument_group("Mimeo Configuration arguments")
        mimeo_config_args.add_argument(
            "-x",
            "--xml-declaration",
            type=str,
            choices=["true", "false"],
            help="overwrite the xml_declaration property")
        mimeo_config_args.add_argument(
            "-i",
            "--indent",
            type=int,
            help="overwrite the indent property")
        mimeo_config_args.add_argument(
            "-o",
            "--output",
            type=str,
            choices=["file", "stdout"],
            help="overwrite the output_details/direction property")
        mimeo_config_args.add_argument(
            "-d",
            "--directory",
            type=str,
            metavar="DIRECTORY_PATH",
            help="overwrite the output_details/directory_path property")
        mimeo_config_args.add_argument(
            "-f",
            "--file",
            type=str,
            metavar="FILE_NAME",
            help="overwrite the output_details/file_name property")

        logging_args = self.add_argument_group("Logging arguments")
        logging_args_excl = logging_args.add_mutually_exclusive_group()
        logging_args_excl.add_argument(
            "--silent",
            action="store_true",
            help="disable INFO logs")
        logging_args_excl.add_argument(
            "--debug",
            action="store_true",
            help="enable DEBUG mode")
        logging_args_excl.add_argument(
            "--fine",
            action="store_true",
            help="enable FINE mode")


def main():
    mimeo_parser = MimeoArgumentParser()
    args = mimeo_parser.parse_args()
    if args.silent:
        logging.getLogger("mimeo").setLevel(logging.WARNING)
    elif args.debug:
        logging.getLogger("mimeo").setLevel(logging.DEBUG)
    elif args.fine:
        logging.getLogger("mimeo").setLevel(logging.FINE)

    logger.info("Starting Mimeo job")
    for path in args.paths:
        logger.info(f"Data generation from Mimeo Config: {path}")
        mimeo_config = get_config(path, args)
        Mimeograph(mimeo_config).produce()


def get_config(config_path, args):
    with open(config_path) as config_file:
        config = json.load(config_file)
        if args.xml_declaration is not None:
            xml_declaration = args.xml_declaration.lower() == "true"
            logger.fine(f"Overwriting xml_declaration to [{xml_declaration}]")
            config[MimeoConfig.XML_DECLARATION_KEY] = xml_declaration
        if args.indent is not None:
            logger.fine(f"Overwriting indent to [{args.indent}]")
            config[MimeoConfig.INDENT_KEY] = args.indent
        if args.output is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_DIRECTION_KEY, args.output)
        if args.directory is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_DIRECTORY_PATH_KEY, args.directory)
        if args.file is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_FILE_NAME_KEY, args.file)
    mimeo_config = MimeoConfig(config)
    logger.debug(f"Mimeo Config: {mimeo_config}")
    return mimeo_config


def customize_output_details(config, key, value):
    if config.get(MimeoConfig.OUTPUT_DETAILS_KEY) is None:
        config[MimeoConfig.OUTPUT_DETAILS_KEY] = {}

    logger.fine(f"Overwriting output details' {key} to [{value}]")
    config[MimeoConfig.OUTPUT_DETAILS_KEY][key] = value


if __name__ == '__main__':
    main()



from mimeo.config.mimeo_config import MimeoConfig
from mimeo.consumers import Consumer, FileConsumer, HttpConsumer, RawConsumer
from mimeo.exceptions import UnsupportedOutputDirection


class ConsumerFactory:

    FILE_DIRECTION = "file"
    STD_OUT_DIRECTION = "stdout"
    HTTP_DIRECTION = "http"

    @staticmethod
    def get_consumer(mimeo_config: MimeoConfig) -> Consumer:
        direction = mimeo_config.output_details.direction
        if direction == ConsumerFactory.STD_OUT_DIRECTION:
            return RawConsumer()
        elif direction == ConsumerFactory.FILE_DIRECTION:
            return FileConsumer(mimeo_config.output_details)
        elif direction == ConsumerFactory.HTTP_DIRECTION:
            return HttpConsumer(mimeo_config.output_details)
        else:
            raise UnsupportedOutputDirection(f"Provided direction [{direction}] is not supported!")



class UnsupportedOutputFormat(Exception):
    pass


class UnsupportedOutputDirection(Exception):
    pass


class UnsupportedAuthMethod(Exception):
    pass


class UnsupportedRequestMethod(Exception):
    pass


class MissingRequiredProperty(Exception):
    pass


class InvalidIndent(Exception):
    pass


class InvalidVars(Exception):
    pass


class IncorrectMimeoModel(Exception):
    pass


class IncorrectMimeoTemplate(Exception):
    pass


class IncorrectMimeoConfig(Exception):
    pass

class ResourceNotFound(Exception):
    pass
